﻿Imports System.Math

Public Class Form1
    ' The point where the user pressed the mouse.
    Private StartPoint As New PointF(-1, -1)

    ' The mouse's current point.
    Private CurrentPoint As New PointF(-1, -1)

    ' Redraw on resize.
    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Me.ResizeRedraw = True
    End Sub

    ' Record the starting point.
    Private Sub Form1_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        StartPoint = e.Location
        CurrentPoint = e.Location
        Me.Refresh()
    End Sub

    ' If the left mouse button is down, move the current point.
    Private Sub Form1_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        If (e.Button = MouseButtons.Left) Then
            CurrentPoint = e.Location
            Me.Refresh()
        End If
    End Sub

    ' Draw the circles, points, and tangent lines.
    Private Sub Form1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias

        ' Draw the first circle.
        Dim radius As Single = Min( _
            Me.ClientSize.Width / 4.0F, _
            Me.ClientSize.Height / 4.0F)
        Dim cx As Single = Me.ClientSize.Width - radius - 20
        Dim cy As Single = Me.ClientSize.Height - radius - 20
        e.Graphics.DrawEllipse(Pens.Blue, _
            cx - radius, cy - radius, _
            2 * radius, 2 * radius)

        If (StartPoint.X < 0) Then Return

        ' Get the second circle's geometry.
        Dim dx As Single = StartPoint.X - CurrentPoint.X
        Dim dy As Single = StartPoint.Y - CurrentPoint.Y
        Dim radius2 As Single = CSng(Sqrt(dx * dx + dy * dy))

        ' Find the tangents.
        Dim outer1_p1, outer1_p2, outer2_p1, outer2_p2 As PointF
        Dim inner1_p1, inner1_p2, inner2_p1, inner2_p2 As PointF
        Dim num_tangents As Integer = FindCircleCircleTangents( _
            New PointF(cx, cy), radius, _
            StartPoint, radius2, _
            outer1_p1, outer1_p2, _
            outer2_p1, outer2_p2, _
            inner1_p1, inner1_p2, _
            inner2_p1, inner2_p2)

        If (num_tangents > 0) Then
            Debug.WriteLine(outer1_p1.ToString())

            ' Draw the outer tangents.
            e.Graphics.DrawLine(Pens.Blue, outer1_p1, outer1_p2)
            e.Graphics.DrawLine(Pens.Blue, outer2_p1, outer2_p2)

            If (num_tangents > 2) Then
                ' Draw the inner tangents.
                e.Graphics.DrawLine(Pens.Red, inner1_p1, inner1_p2)
                e.Graphics.DrawLine(Pens.Red, inner2_p1, inner2_p2)
            End If
        End If

        ' Draw the second circle.
        e.Graphics.DrawEllipse(Pens.Red, _
            StartPoint.X - radius2, StartPoint.Y - radius2, _
            2 * radius2, 2 * radius2)
    End Sub

    ' Find the tangent points for these two circles.
    ' Return the number of tangents: 4, 2, or 0.
    Private Function FindCircleCircleTangents( _
        ByVal c1 As PointF, ByVal radius1 As Single, ByVal c2 As PointF, ByVal radius2 As Single, _
        ByRef outer1_p1 As PointF, ByRef outer1_p2 As PointF, _
        ByRef outer2_p1 As PointF, ByRef outer2_p2 As PointF, _
        ByRef inner1_p1 As PointF, ByRef inner1_p2 As PointF, _
        ByRef inner2_p1 As PointF, ByRef inner2_p2 As PointF) _
    As Integer
        ' Make sure radius1 <= radius2.
        If (radius1 > radius2) Then
            ' Call this method switching the circles.
            Return FindCircleCircleTangents( _
                c2, radius2, c1, radius1, _
                outer1_p2, outer1_p1, _
                outer2_p2, outer2_p1, _
                inner1_p2, inner1_p1, _
                inner2_p2, inner2_p1)
        End If

        ' Initialize the return values in case
        ' some tangents are missing.
        outer1_p1 = New PointF(-1, -1)
        outer1_p2 = New PointF(-1, -1)
        outer2_p1 = New PointF(-1, -1)
        outer2_p2 = New PointF(-1, -1)
        inner1_p1 = New PointF(-1, -1)
        inner1_p2 = New PointF(-1, -1)
        inner2_p1 = New PointF(-1, -1)
        inner2_p2 = New PointF(-1, -1)

        ' ***************************
        ' * Find the outer tangents *
        ' ***************************
        Dim radius2a As Single = radius2 - radius1
        If (Not FindTangents(c2, radius2a, c1, outer1_p2, outer2_p2)) Then
            ' There are no tangents.
            Return 0
        End If

        ' Get the vector perpendicular to the
        ' first tangent with length radius1.
        Dim v1x As Single = -(outer1_p2.Y - c1.Y)
        Dim v1y As Single = outer1_p2.X - c1.X
        Dim v1_length As Single = CSng(Sqrt(v1x * v1x + v1y * v1y))
        v1x *= radius1 / v1_length
        v1y *= radius1 / v1_length
        ' Offset the tangent vector's points.
        outer1_p1 = New PointF(c1.X + v1x, c1.Y + v1y)
        outer1_p2 = New PointF(outer1_p2.X + v1x, outer1_p2.Y + v1y)

        ' Get the vector perpendicular to the
        ' second tangent with length radius1.
        Dim v2x As Single = outer2_p2.Y - c1.Y
        Dim v2y As Single = -(outer2_p2.X - c1.X)
        Dim v2_length As Single = CSng(Sqrt(v2x * v2x + v2y * v2y))
        v2x *= radius1 / v2_length
        v2y *= radius1 / v2_length
        ' Offset the tangent vector's points.
        outer2_p1 = New PointF(c1.X + v2x, c1.Y + v2y)
        outer2_p2 = New PointF(outer2_p2.X + v2x, outer2_p2.Y + v2y)

        ' If the circles intersect, then there are no inner tangents.
        Dim dx As Single = c2.X - c1.X
        Dim dy As Single = c2.Y - c1.Y
        Dim dist As Double = Sqrt(dx * dx + dy * dy)
        If (dist <= radius1 + radius2) Then Return 2

        ' ***************************
        ' * Find the inner tangents *
        ' ***************************
        Dim radius1a As Single = radius1 + radius2
        FindTangents(c1, radius1a, c2, inner1_p2, inner2_p2)

        ' Get the vector perpendicular to the
        ' first tangent with length radius2.
        v1x = inner1_p2.Y - c2.Y
        v1y = -(inner1_p2.X - c2.X)
        v1_length = CSng(Sqrt(v1x * v1x + v1y * v1y))
        v1x *= radius2 / v1_length
        v1y *= radius2 / v1_length
        ' Offset the tangent vector's points.
        inner1_p1 = New PointF(c2.X + v1x, c2.Y + v1y)
        inner1_p2 = New PointF(inner1_p2.X + v1x, inner1_p2.Y + v1y)

        ' Get the vector perpendicular to the
        ' second tangent with length radius2.
        v2x = -(inner2_p2.Y - c2.Y)
        v2y = inner2_p2.X - c2.X
        v2_length = CSng(Sqrt(v2x * v2x + v2y * v2y))
        v2x *= radius2 / v2_length
        v2y *= radius2 / v2_length
        ' Offset the tangent vector's points.
        inner2_p1 = New PointF(c2.X + v2x, c2.Y + v2y)
        inner2_p2 = New PointF(inner2_p2.X + v2x, inner2_p2.Y + v2y)

        Return 4
    End Function

    ' Find the tangent points for this circle and external point.
    ' Return true if we find the tangents, false if the point is
    ' inside the circle.
    Private Function FindTangents(ByVal center As PointF, ByVal radius As Single, _
        ByVal external_point As PointF, ByRef pt1 As PointF, ByRef pt2 As PointF) As Boolean

        ' Find the distance squared from the
        ' external point to the circle's center.
        Dim dx As Double = center.X - external_point.X
        Dim dy As Double = center.Y - external_point.Y
        Dim D_squared As Double = dx * dx + dy * dy
        If (D_squared < radius * radius) Then
            pt1 = New PointF(-1, -1)
            pt2 = New PointF(-1, -1)
            Return False
        End If

        ' Find the distance from the external point
        ' to the tangent points.
        Dim L As Double = Sqrt(D_squared - radius * radius)

        ' Find the points of intersection between
        ' the original circle and the circle with
        ' center external_point and radius dist.
        FindCircleCircleIntersections( _
            center.X, center.Y, radius, _
            external_point.X, external_point.Y, CSng(L), _
            pt1, pt2)

        Return True
    End Function

    ' Find the points where the two circles intersect.
    Private Function FindCircleCircleIntersections( _
     ByVal cx0 As Single, ByVal cy0 As Single, ByVal radius0 As _
         Single, _
     ByVal cx1 As Single, ByVal cy1 As Single, ByVal radius1 As _
         Single, _
     ByRef intersection1 As PointF, ByRef intersection2 As  _
         PointF) As Integer
        ' Find the distance between the centers.
        Dim dx As Single = cx0 - cx1
        Dim dy As Single = cy0 - cy1
        Dim dist As Double = Sqrt(dx * dx + dy * dy)

        ' See how many solutions there are.
        If (dist > radius0 + radius1) Then
            ' No solutions, the circles are too far apart.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        ElseIf (dist < Abs(radius0 - radius1)) Then
            ' No solutions, one circle contains the other.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        ElseIf ((dist = 0) AndAlso (radius0 = radius1)) Then
            ' No solutions, the circles coincide.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        Else
            ' Find a and h.
            Dim a As Double = (radius0 * radius0 - _
                    radius1 * radius1 + dist * dist) / (2 * _
                        dist)
            Dim h As Double = Sqrt(radius0 * radius0 - a * _
                a)

            ' Find P2.
            Dim cx2 As Double = cx0 + a * (cx1 - cx0) / dist
            Dim cy2 As Double = cy0 + a * (cy1 - cy0) / dist

            ' Get the points P3.
            intersection1 = New PointF( _
                CSng(cx2 + h * (cy1 - cy0) / dist), _
                CSng(cy2 - h * (cx1 - cx0) / dist))
            intersection2 = New PointF( _
                CSng(cx2 - h * (cy1 - cy0) / dist), _
                CSng(cy2 + h * (cx1 - cx0) / dist))

            ' See if we have 1 or 2 solutions.
            If (dist = radius0 + radius1) Then Return 1
            Return 2
        End If
    End Function
End Class
